from huey import crontab  # type: ignore
from huey.contrib.djhuey import db_periodic_task  # type: ignore
from django.core.management import call_command


@db_periodic_task(crontab(minute=5))
def kill_unresponsive_tasks_task():
    call_command("kill_unresponsive_tasks")



from django.contrib import admin
from django.contrib.auth import (
    get_user_model
)
from django.utils.translation import gettext as _
from kbackgroundtask.models import (
    BackgroundTask,
)
from jutil.admin import ModelAdminBase
from django.contrib.admin import SimpleListFilter
from django.utils.timezone import now
from datetime import timedelta
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from jutil.format import format_timedelta
from django.http import HttpRequest
from django.core.management import call_command
from django.contrib.messages import add_message
from django.core.checks import messages
from django.shortcuts import redirect

User = get_user_model()


class BackgroundTaskStateFilter(SimpleListFilter):
    title = _("state")
    parameter_name = "state"

    def lookups(self, request, model_admin):
        opts = [
            ("0", _("ok")),
            ("1", _("non-completed")),
            ("2-1", _("not executed over 24 hours")),
            ("2-7", _("not executed over 7 days")),
            ("2-31", _("not executed over 31 days")),
        ]
        for obj in list(BackgroundTask.objects.all().distinct("signal")):
            assert isinstance(obj, BackgroundTask)
            opts.append((obj.signal, _(obj.signal)))
        return opts

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            if val == "0":
                return queryset.exclude(signal="error")
            if val == "1":
                return queryset.filter(
                    signal__in=BackgroundTask.UNFINISHED_TASK_SIGNALS
                )
            if val.startswith("2-"):
                days = int(val.split("-")[1])
                active_tasks = list(
                    BackgroundTask.objects.filter(
                        created__gt=now() - timedelta(days=days)
                    )
                    .distinct("name")
                    .values_list("name", flat=True)
                )
                return queryset.exclude(name__in=active_tasks)
            return queryset.filter(signal=val)
        return queryset


class BackgroundTaskAdmin(ModelAdminBase):
    actions = ()
    date_hierarchy = "created"
    list_filter = (
        BackgroundTaskStateFilter,
        "name",
    )
    search_fields = ("=task_id",)
    list_display = (
        "id",
        "created",
        "signal_localized",
        "name_link",
        "admin_runtime",
        "admin_error_brief",
    )
    fields = (
        "id",
        "name",
        "task_id",
        "created",
        "last_modified",
        "signal_localized",
        "complete",
        "executing",
        "locked",
        "retrying",
        "scheduled",
        "canceled",
        "revoked",
        "failed",
        "error",
        "admin_runtime",
    )
    readonly_fields = fields

    def admin_error_brief(self, obj):
        assert isinstance(obj, BackgroundTask)
        err = obj.error_brief
        font_color = "green" if obj.is_newer_ok() else "red"
        return format_html('<span style="color:{}">{}</span>', font_color, err)

    admin_error_brief.short_description = _("error")  # type: ignore

    def signal_localized(self, obj) -> str:
        assert isinstance(obj, BackgroundTask)
        return _(obj.signal)

    signal_localized.short_description = _("signal")  # type: ignore
    signal_localized.admin_order_field = "signal"  # type: ignore

    def admin_runtime(self, obj):
        assert isinstance(obj, BackgroundTask)
        dt = obj.runtime
        if dt is not None:
            return format_timedelta(dt) or "<1s"
        return ""

    admin_runtime.short_description = _("runtime")  # type: ignore

    def name_link(self, obj):
        assert isinstance(obj, BackgroundTask)
        return mark_safe(
            f"<a href='/admin/backgroundtask/backgroundtask/?name={obj.name}'>{obj.name}</a>"
        )

    name_link.short_description = _("name")  # type: ignore
    name_link.admin_order_field = "name"  # type: ignore

    def task_link(self, obj):
        assert isinstance(obj, BackgroundTask)
        return mark_safe(
            f"<a href='/admin/backgroundtask/backgroundtask/?q={obj.task_id}'>{obj.task_id}</a>"
        )

    task_link.short_description = _("task")  # type: ignore
    task_link.admin_order_field = "task_id"  # type: ignore

    def changelist_view(self, request: HttpRequest, extra_context=None):
        if request.GET.get("flush-redis", "") == "1":
            call_command("redis_flushdb", force=True)
            add_message(request, messages.INFO, "Background task Redis locks flushed")
            return redirect(request.META["HTTP_REFERER"])
        if request.GET.get("reset-error-if-newer-ok", "") == "1":
            for bt in list(BackgroundTask.objects.filter_unfinished()):
                assert isinstance(bt, BackgroundTask)
                if bt.is_newer_ok():
                    add_message(request, messages.INFO, f"{bt} deleted")
                    bt.delete()
            add_message(
                request,
                messages.INFO,
                "Errors reset for tasks which have newer complete runs",
            )
            return redirect(request.META["HTTP_REFERER"])
        return super().changelist_view(request, extra_context)


admin.site.register(BackgroundTask, BackgroundTaskAdmin)



import logging
from typing import Optional, Sequence, Union, Tuple, List
import re
from django.conf import settings
from django.template.loader import render_to_string
from jutil.email import (
    send_email as send_email_impl,
)

logger = logging.getLogger(__name__)


def send_email(  # noqa
    recipients: Sequence[Union[str, Tuple[str, str]]],
    subject: str,
    text: str = "",
    html: str = "",
    sender: Union[str, Tuple[str, str]] = "",
    files: Optional[Sequence[str]] = None,
    files_content: Optional[Sequence[Union[Tuple[str, bytes], Tuple[str, bytes, str]]]] = None,
    cc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    bcc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    exceptions: bool = False,
):

    if not settings.EMAIL_SENDING_ENABLED:
        logger.info(
            "Email sending disabled (EMAIL_SENDING_ENABLED=%s, DEBUG=%s):",
            settings.EMAIL_SENDING_ENABLED,
            settings.DEBUG,
        )
        logger.info("To: %s", recipients)
        logger.info("Subject: %s", subject)
        if sender:
            logger.info("Sender: %s", sender)
        if cc_recipients:
            logger.info("Cc: %s", cc_recipients)
        if bcc_recipients:
            logger.info("Bcc: %s", bcc_recipients)
        logger.info("%s", html or text)
        return

    return send_email_impl(
        recipients=recipients,
        subject=subject,
        text=text,
        html=html,
        sender=sender,
        files=files,
        files_content=files_content,
        cc_recipients=cc_recipients,
        bcc_recipients=bcc_recipients,
        exceptions=exceptions,
    )


def send_email_with_template(  # noqa
    recipients: Sequence[Union[str, Tuple[str, str]]],
    subject: str,
    html_template: Optional[str] = None,
    body_params: Optional[dict] = None,
    sender: Union[str, Tuple[str, str]] = "",
    files: Optional[Sequence[str]] = None,
    cc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    bcc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    exceptions: bool = False,
):
    body_params = body_params if body_params is not None else {}
    if html_template is not None:
        html = render_to_string(html_template, body_params)

    send_email(
        recipients=recipients,
        subject=subject,
        html=html,
        sender=sender,
        files=files,
        cc_recipients=cc_recipients,
        bcc_recipients=bcc_recipients,
        exceptions=exceptions,
    )




def csv_to_list(value_list_str: str) -> List[str]:
    """
    Splits string by 1) ' ' 2) ';' 3) ','
    and returns list of strings.
    :param value_list_str: E.g. "john@gmail.com, henry@gmail.com"
    :return: List of emails
    """
    value_list_str = value_list_str.replace(",", " ")
    value_list_str = value_list_str.replace(";", " ")
    value_list_str = re.sub(r"\s+", " ", value_list_str)
    vals = []
    for val in value_list_str.split(" "):
        val = val.strip()
        if val:
            vals.append(val)
    return vals


# Generated by Django 3.2.7 on 2021-10-14 08:14

from django.db import migrations, models
import django.utils.timezone
import jutil.modelfields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BackgroundTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', jutil.modelfields.SafeCharField(db_index=True, max_length=64, verbose_name='name')),
                ('task_id', jutil.modelfields.SafeCharField(db_index=True, max_length=64, verbose_name='task')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, verbose_name='created')),
                ('last_modified', models.DateTimeField(auto_now=True, db_index=True, verbose_name='last modified')),
                ('signal', jutil.modelfields.SafeCharField(db_index=True, max_length=64, verbose_name='signal')),
                ('executing', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='executing')),
                ('complete', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='complete')),
                ('locked', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='locked')),
                ('scheduled', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='scheduled')),
                ('retrying', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='retrying')),
                ('canceled', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='canceled')),
                ('revoked', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='revoked')),
                ('failed', models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name='failed')),
                ('error', jutil.modelfields.SafeTextField(blank=True, verbose_name='error')),
            ],
            options={
                'verbose_name': 'background task',
                'verbose_name_plural': 'background tasks',
            },
        ),
    ]



import logging
from huey.contrib.djhuey import HUEY as huey  # type: ignore
from django.utils.timezone import now
import traceback
from kbackgroundtask.models import BackgroundTask

TASKS_TO_IGNORE = {"run_check"}


logger = logging.getLogger(__name__)


@huey.signal()
def all_signal_handler(signal, task, exc=None):
    """
    task.__dict__ == {
        'args': (),
        'eta': None,
        'id': '01d685e1-aefe-4fbd-b37c-8700392781d2',
        'kwargs': {},
        'name': 'dummy_db_periodic_task_1',
        'on_complete': None,
        'on_error': None,
        'priority': None,
        'retries': 0,
        'retry_delay': 0,
        'revoke_id': 'r:01d685e1-aefe-4fbd-b37c-8700392781d2'
    }
    """
    time_now = now()
    err = ""
    if exc:
        try:
            err = traceback.format_exc()
        except Exception:
            err = str(exc)
    if err or task.name not in TASKS_TO_IGNORE:
        if signal == "executing":
            logger.info(
                "Executing task id=%s: %s(*%s, **%s)",
                task.id,
                task.name,
                task.args,
                task.kwargs,
            )
        update_fields = ["signal", "last_modified"]
        obj = BackgroundTask.objects.get_or_create(
            task_id=task.id, defaults={"name": task.name}
        )[0]
        assert isinstance(obj, BackgroundTask)
        obj.signal = signal
        if not hasattr(obj, signal):
            logger.error("Object %s does not have member %s (signal)", obj, signal)
        else:
            setattr(obj, signal, time_now)
            update_fields.append(signal)
        obj.last_modified = time_now
        if err:
            obj.error = err
            obj.failed = time_now
            update_fields.append("error")
            update_fields.append("failed")
        obj.save(update_fields=update_fields)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

class BackgroundTaskConfig(AppConfig):
    name = "kbackgroundtask"
    verbose_name = _('Background Task Management')
    # default_auto_field = "django.db.models.AutoField"
    def ready(self):
        from . import signals as signals_init  # noqa



from typing import Optional, Sequence

from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from jutil.modelfields import SafeCharField, SafeTextField
from datetime import datetime, timedelta
import logging
from time import sleep

logger = logging.getLogger(__name__)


class BackgroundTaskManager(models.Manager):
    def filter_unfinished(self):
        return self.filter(signal__in=BackgroundTask.UNFINISHED_TASK_SIGNALS)

    def filter_executing(self):
        return self.filter(signal="executing")

    def wait_tasks(
        self,
        wait_s: int = 0,
        signals: Optional[Sequence[str]] = None,
        poll_s: float = 1.0,
        verbose: bool = True,
    ):
        """
        Waits for tasks to finish specified signal(s) state.
        :param wait_s: Max seconds to wait. Default: 0
        :param signals: Signals to wait. Default: ['executing']
        :param poll_s: Polling interval seconds. Default: 1.0
        :param verbose: Print logger.info output about wait.
        :return: BackgroundTask in specified signal or None if all finished
        """
        if signals is None:
            signals = ["executing"]
        task: Optional[BackgroundTask] = None
        time_now = now()
        end_time = time_now + timedelta(seconds=wait_s) if wait_s else time_now
        while time_now <= end_time:
            task = self.all().filter(signal__in=signals).first()
            if task is None:
                break
            if verbose:
                logger.info(
                    "Waiting %s tasks %s: %s", signals, end_time - time_now, task
                )
            sleep(poll_s)
            time_now = now()
        return task


class BackgroundTask(models.Model):
    ACTIVE_TASK_SIGNALS = ["executing", "scheduled", "retrying"]
    UNFINISHED_TASK_SIGNALS = ACTIVE_TASK_SIGNALS + ["error"]

    objects = BackgroundTaskManager()
    name = SafeCharField(_("name"), max_length=64, db_index=True)
    task_id = SafeCharField(_("task"), max_length=64, db_index=True)
    created = models.DateTimeField(_("created"), default=now, blank=True, db_index=True)
    last_modified = models.DateTimeField(
        verbose_name=_("last modified"),
        auto_now=True,
        db_index=True,
        editable=False,
        blank=True,
    )
    signal = SafeCharField(_("signal"), max_length=64, db_index=True)
    executing = models.DateTimeField(
        _("executing"), default=None, blank=True, null=True, db_index=True
    )
    complete = models.DateTimeField(
        _("complete"), default=None, blank=True, null=True, db_index=True
    )
    locked = models.DateTimeField(
        _("locked"), default=None, blank=True, null=True, db_index=True
    )
    scheduled = models.DateTimeField(
        _("scheduled"), default=None, blank=True, null=True, db_index=True
    )
    retrying = models.DateTimeField(
        _("retrying"), default=None, blank=True, null=True, db_index=True
    )
    canceled = models.DateTimeField(
        _("canceled"), default=None, blank=True, null=True, db_index=True
    )
    revoked = models.DateTimeField(
        _("revoked"), default=None, blank=True, null=True, db_index=True
    )
    failed = models.DateTimeField(
        _("failed"), default=None, blank=True, null=True, db_index=True
    )
    error = SafeTextField(_("error"), blank=True)

    class Meta:
        verbose_name = _("background task")
        verbose_name_plural = _("background tasks")

    def __str__(self):
        return f"{self.signal} {self.name} ({self.task_id})"

    def is_newer_ok(self) -> bool:
        """
        Returns True if there is newer task with same name that has succeed.
        """
        return (
            bool(self.complete)
            or BackgroundTask.objects.filter(
                name=self.name, complete__gt=self.created
            ).exists()
        )

    @property
    def error_brief(self) -> str:
        if self.error:
            lines = str(self.error).strip().split("\n")
            return lines.pop()
        return ""

    error_brief.fget.short_description = _("error")  # type: ignore

    @property
    def end_time(self) -> Optional[datetime]:
        max_t: Optional[datetime] = None
        for t in [self.complete, self.locked, self.revoked, self.canceled, self.failed]:
            if max_t is None or t is not None and t > max_t:
                max_t = t
        return max_t

    end_time.fget.short_description = _("end time")  # type: ignore

    @property
    def runtime(self) -> Optional[timedelta]:
        min_t, max_t = self.executing, self.end_time
        if max_t is None:
            max_t = now()
        return max_t - min_t if min_t is not None and max_t is not None else None

    runtime.fget.short_description = _("runtime")  # type: ignore



import json
import logging
from functools import lru_cache
from typing import List, Dict, Union, Optional, Any
import sendgrid  # type: ignore
from django.conf import settings


logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def get_sendgrid_client() -> sendgrid.SendGridAPIClient:
    return sendgrid.SendGridAPIClient(api_key=settings.EMAIL_SENDGRID_API_KEY)  # type: ignore


def get_email_bounce_list() -> List[Dict[str, Union[int, str]]]:
    sg = get_sendgrid_client()
    response = sg.client.suppression.bounces.get()
    if response.status_code >= 300:
        logger.warning("get_email_bounce_list failed:")
        logger.warning(response.status_code)
        logger.warning(response.body)
        logger.warning(response.headers)
        raise Exception(
            f"SendGrid GET suppression/bounces failed: {response.body.decode()}"
        )
    return json.loads(response.body)


def delete_email_bounce_list(
    delete_all: bool = True, emails: Optional[List[str]] = None
):
    sg = get_sendgrid_client()
    data: Dict[str, Any] = {
        "delete_all": delete_all,
    }
    if emails:
        data["emails"] = emails
    response = sg.client.suppression.bounces.delete(data)
    if response.status_code >= 300:
        logger.warning("delete_email_bounce_list failed:")
        logger.warning(response.status_code)
        logger.warning(response.body)
        logger.warning(response.headers)
        raise Exception(
            f"SendGrid DELETE suppression/bounces failed: {response.body.decode()}"
        )



import logging
from datetime import timedelta
from typing import List
from kcommunication.helpers import send_email
from kcommunication.helpers import csv_to_list
from django.conf import settings
from django.core.management.base import CommandParser
from django.utils.html import strip_tags
from django.utils.timezone import now
from kbackgroundtask.models import BackgroundTask
from jutil.command import SafeCommand
from project.redis_services import get_redis_instance


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = "Kills unresponsive tasks"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("--email", type=str)
        parser.add_argument("--no-email", action="store_true")
        parser.add_argument("--max-runtime-minutes", type=int, default=30)
        parser.add_argument("--max-consecutive-locks", type=int, default=2)

    def do(self, *args, **options):  # noqa
        time_now = now()
        emails = csv_to_list(options["email"]) if options["email"] else settings.ADMINS

        max_minutes = options["max_runtime_minutes"]
        old = time_now - timedelta(minutes=max_minutes)

        # list tasks stuck in 'executing' state
        exec_tasks: List[BackgroundTask] = list(
            BackgroundTask.objects.filter(created__lt=old, signal__in=["executing"])
        )

        # list tasks which are stuck in locked state
        locked_tasks: List[BackgroundTask] = []
        max_locks = options["max_consecutive_locks"]
        for bt in list(
            BackgroundTask.objects.filter(created__gt=old, signal="locked")
            .order_by("name")
            .distinct("name")
        ):
            assert isinstance(bt, BackgroundTask)
            qs: List[BackgroundTask] = list(
                BackgroundTask.objects.filter(name=bt.name).order_by("-id")[:max_locks]
            )
            qs_locked = [bt for bt in qs if bt.signal == "locked"]
            if len(qs_locked) == max_locks:
                bt = qs[0]
                locked_tasks.append(bt)

        # if we have executing tasks running too long time or stuck in locked state, flush Redis locks
        if exec_tasks or locked_tasks:
            redis = get_redis_instance()
            redis.flushdb()
            logger.info("Redis flushed")
            for bt in exec_tasks:
                assert isinstance(bt, BackgroundTask)
                bt.error = (
                    f"Set as failed since runtime {bt.runtime} exceeds {max_minutes}m. Locks flushed."
                )
                bt.signal = "error"
                bt.last_modified = time_now
                bt.save()
                logger.info("%s: %s", bt, bt.error)
            for bt in locked_tasks:
                assert isinstance(bt, BackgroundTask)
                bt.error = "Set as canceled since task looks stuck in locked state. Locks flushed."
                bt.signal = "canceled"
                bt.last_modified = time_now
                bt.save()
                logger.info("%s: %s", bt, bt.error)

            print("To:", emails)
            subject = f"{settings.SITE_NAME} background tasks unresponsive"
            print("Subject:", subject)
            list_url = settings.API_URL + "/admin/customers/backgroundtask/?state=1"
            html = f'<p><a href="{list_url}">{list_url}</a></p>\n'
            html += '<p>Tasks stuck in "executing" state (now released):</p><ul>\n'
            for bt in exec_tasks:
                assert isinstance(bt, BackgroundTask)
                html += f"  <li>{bt.name} ({bt.signal} {time_now - bt.created})</li>\n"
            if not exec_tasks:
                html += "<li>(no tasks)</li>"
            html += "</ul>\n"
            html += '<p>Tasks stuck in "locked" state (now released):</p><ul>\n'
            for bt in locked_tasks:
                assert isinstance(bt, BackgroundTask)
                html += f"  <li>{bt.name} ({bt.signal} {time_now - bt.created})</li>\n"
            if not locked_tasks:
                html += "<li>(no tasks)</li>"
            html += "</ul>\n"
            text = strip_tags(html.replace("<br>", "\n"))
            print(html)
            if not options["no_email"]:
                if exec_tasks:
                    send_email(emails, subject, text, html)
                else:
                    print("Not sending email since no tasks in error state")



import logging
from typing import Union, Tuple, Sequence, Optional
from huey import crontab  # type: ignore
from huey.contrib.djhuey import db_periodic_task, task  # type: ignore
from kcommunication.helpers import send_email, send_email_with_template
from kcommunication.sendgrid import delete_email_bounce_list

logger = logging.getLogger(__name__)

@task(retries=3, retry_delay=60, priority=10)
def send_email_with_template_task(email, subject, template, body):
    send_email_with_template(email, subject, template, body)


@task(retries=3, retry_delay=60, priority=10)
def send_email_task(  # noqa
    recipients: Sequence[Union[str, Tuple[str, str]]],
    subject: str,
    text: str = "",
    html: str = "",
    sender: Union[str, Tuple[str, str]] = "",
    files: Optional[Sequence[str]] = None,
    cc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    bcc_recipients: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
    exceptions: bool = False,
):
    send_email(
        recipients=recipients,
        subject=subject,
        text=text,
        html=html,
        sender=sender,
        files=files,
        cc_recipients=cc_recipients,
        bcc_recipients=bcc_recipients,
        exceptions=exceptions,
    )


@db_periodic_task(crontab(minute=0, hour=4))
def sendgrid_delete_bounce_list_daily_task():
    delete_email_bounce_list()



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages  # type: ignore


def parse_requirements(filename, session=False):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


install_requires = parse_requirements("requirements.txt", session=False)

setup(
    name="django-kcommunication",
    version="1.2.2",
    author=u"Tuan Bach Van",
    author_email="tuan@kajala.com",
    packages=find_packages(exclude=["project", "venv"]),
    include_package_data=True,
    url="https://github.com/kajalagroup/communication",
    license="MIT licence, see LICENCE.txt",
    description="Communication helpers and background task management for Huey queue.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    zip_safe=False,
    install_requires=install_requires,
)
